using System;

namespace MyRM
{
 /// <summary>
 /// class MyRM implements TP.RM
 /// </summary>
 public class MyRM : System.MarshalByRefObject, TP.RM
 {
  MyLM lockManager;
  TP.TM transactionManager;
  System.Collections.Hashtable Items;
  System.Collections.Hashtable Reservations;
		
  public MyRM()
  {
      this.lockManager = new MyLM();

	  /*  Don't need this in the beginning */
      /*  this.transactionManager = (TP.TM) System.Activator.GetObject( typeof( MyTM.MyTM ), "http://localhost:8081/TM.soap");
		  TP.Transaction tid = transactionManager.Start();
		  transactionManager.Abort( tid );
      */
      
      Items = new System.Collections.Hashtable();
      Reservations = new System.Collections.Hashtable();
  }


  class Reservation
  {
      public TP.Customer Holder;
      public System.Collections.ArrayList Items;

	  public Reservation(TP.Customer c, TP.Item i, int price)
	  {
		this.Holder = c;
		Items = new System.Collections.ArrayList();
		Items.Add(new ItemReservation(i, price));
	  }
  }


  class ItemReservation
  {
      public TP.Item ThisItem;
      public int Price;

	  public ItemReservation(TP.Item i, int price)
	  {
		this.ThisItem = i;
		Price = price;
	  }
  }


  class ItemInstance
  {
    string Name
	{
	  get
	  {
	    return i.Name;
	  }
	}

    TP.Item i;
    public int Count;
    public int CurrentPrice;

	public ItemInstance(TP.Item i, int count, int price)
	{
	  this.i = i;
	  this.Count = count;
	  this.CurrentPrice = price;
	}
  }


  static void Main(string [] args)
  {
      if (args.Length<1)
	  {
		Console.WriteLine("Usage: MyRM <rmName> [port_number]");
		return;
	  }

	  string port_num;
	  if (args.Length<2)
	  {
		  // default port number
		  port_num="8081";
	  }
	  else
		{
		  port_num=args[1];
		}

      System.Console.WriteLine(string.Format( "Starting resource manager for {0} on port {1}", args[0], port_num));

	  // Wait for Transaction manager to come up
      /* System.Threading.Thread.Sleep(4000); */

      System.Runtime.Remoting.Channels.Http.HttpChannel channel 
		  = new System.Runtime.Remoting.Channels.Http.HttpChannel(Convert.ToInt32(port_num));

      System.Runtime.Remoting.Channels.ChannelServices.RegisterChannel(channel);

      System.Runtime.Remoting.RemotingConfiguration.RegisterWellKnownServiceType
	  (		Type.GetType("MyRM.MyRM")									// Assembly name
			,	"RM.soap"												// URI
			,	System.Runtime.Remoting.WellKnownObjectMode.Singleton	// Instancing mode
	  );

      while( true )
		System.Threading.Thread.Sleep(100000);
  }


  // Call to TM to enlist for distributed transaction
  public void Enlist(TP.Transaction context)
  {
    // transactionManager.Enlist(context);
  }

  public void Commit(TP.Transaction context)
  {
    // transactionManager.Commit(context);
  }

  public void Abort(TP.Transaction context)
  {
    // transactionManager.Abort(context);
  }
    
  // Need to add code
  // This method adds a resource to the available ones
  public bool Add(TP.Transaction context, TP.Item i, int count, int price)
  {
      ItemInstance ii = (ItemInstance)Items[i.Name];

      if(ii != null)
	  {
		Console.WriteLine("RM: Add: item is not null {0} ",i);
		ii.Count += count;
		ii.CurrentPrice = price;
	  }
      else
		{
			Console.WriteLine("RM: Add: item is null {0} ",i);
			Items.Add(i.Name, new ItemInstance(i, count, price));
			
			// Need to add locking code here
		}

	return true;
  }


  // Need to add code
  public bool Delete(TP.Transaction context, TP.Item i)
  {
      // Need to add locking code here

      Items.Remove(i.Name);
      return true;
  }

		
  /// <summary>
  /*  NEED TO ADD CODE For STEP 2
		Calling shutdown causes RM to exit gracefully.
		This means, it waits for all the existing transactions 
		to end and enlist requests for new transactions are refused. 
		If any of the existing transactions blocks forever, 
		a retry/timeout mechanism is used to exit.
		No recovery is done on startup */
  /// </summary>
  public void Shutdown()
  {
  }

  /// <summary>
  /*    Exit after the specified number of disk writes.
		Support for this method requires a wrapper around _write_ system
		call that decrements the counter set by this method.

  		This counter should be set by default to 0, which implies that the wrapper
		will do nothing.  If it is non-zero, the wrapper should decrement
		the counter, see if it becomes zero, and if so, call exit(), otherwise
		continue to write. */
  /// </summary>
  public void SelfDestruct(int diskWritesToWait)
  {
  }
   
  /// <summary>
  /*   Need to add code here
	   returns the amount available for the specified item type */
  /// </summary>
  public int Query(TP.Transaction context, TP.Item i)
  {
      Console.WriteLine("RM: Query");
      ItemInstance ii = (ItemInstance)Items[i.Name];
  
	  if(ii == null)
	  {
		  return 0;
	  }
	  else
	  {
		  return ii.Count;
	  }
  }


  // <summary>
  /* Need to add code here
	 returns the price for the specified item type */
  // </summary>
  public int QueryPrice(TP.Transaction xId, TP.Item i)
  {
      return ((ItemInstance)Items[i.Name]).CurrentPrice;
  }


  /// <summary>
  /*   Need to add code here for locking: STEP 2
		Query the bill for a customer. */
  ///</summary>
  /// <param name="context"></param>
  /// <param name="c">customer to be queried</param>
  /// <returns>Items list for the customer</returns>
  public string QueryCustomerInfo(TP.Transaction context, TP.Customer c)
  {
      string order;
      order = "";

      Reservation r = (Reservation)Reservations[c.Id];

	  // Loop over reservations for the passed in customer
      foreach(ItemReservation rr in r.Items)
	  {
		  if(order.Length > 0)
		  {
			  order += ",";
		  }

		order += rr.ThisItem.Name;
	  }

    return order;
  }


  /// <summary>
  /*   Need to add code here for locking: STEP 2
	   Query the total for all reservations for a customer. */
  ///</summary>
  /// <param name="c">customer to be queried</param>
  /// <returns>Total for all reservations for the customer</returns>
  public int QueryCustomer(TP.Transaction context, TP.Customer c)
  {
      int bill=0;
      
      Reservation r = (Reservation)Reservations[c.Id];

      foreach(ItemReservation rr in r.Items)
	  {
		 bill+= rr.Price;
	  }

      return bill ;
  }
	
			
  /// <summary>
  /// Need to add code here
  /// </summary>
  /// <param name="context"></param>
  /// <param name="c">Customer</param>
  /// <param name="i">The item to reserve</param>
  public bool Reserve(TP.Transaction context, TP.Customer c, TP.Item i)
  {
      ItemInstance ii = (ItemInstance)Items[i.Name];
	  if(ii == null)
	  {
		  throw new System.Exception(string.Format("Item {0} does not exists", i.Name));
	  }

	  if(ii.Count > 0)
	  {
		  Reservation r = (Reservation)Reservations[c.Id];
		  if( r == null ) 
		  {
			  r = new Reservation(c, i, ii.CurrentPrice);
			  Reservations.Add(c.Id, r);
		  }
		  else
		  {
			  r.Items.Add(new ItemReservation(i,ii.CurrentPrice));
		  }

		  ii.Count -= 1;
		  return true;
	  }
	  else
	  {
		  throw new System.Exception("No more items available");
	  }
  }


  /// <summary>
  /// Need to add code here
  /// </summary>
  /// <param name="context"></param>
  /// <param name="c">Customer</param>
  public void DropReservations(TP.Transaction context, TP.Customer c)
  {
      Reservation r = (Reservation)Reservations[c.Id];

	  // Loop over reserved items in the reservation and make them available
	  foreach(ItemReservation rr in r.Items)
	  {
		  ((ItemInstance)Items[rr.ThisItem]).Count += 1;
	  }

	  Reservations.Remove(c.Id);
  }
 }
}
